import numpy as np
import pandas as pd
from dataclasses import dataclass
from sklearn.base import BaseEstimator, TransformerMixin


@dataclass
class EventCountersPostproc(BaseEstimator, TransformerMixin):
    """
    Для дополнительной пост-обработки выходных счетчиков трансформера
    EventCounters.

    *_event_count, *_visited_count, *_ignored_count делим на значения
    из day_of_life поле

    *_visit_ratio, *_ignore_ratio сглаживаем с использованием
    Laplace-отношений (+k) / (+k+m)

    *_consecutive_ignores обрезаем сверху.

    Parameters
    -----------
    event_name: str
        Имя события, лежащее в префиксе признаков.
    days_col: str, default  "days_of_life"
        Колонка с временем жизни инживидуального временного ряда.
    clip_consec: int, default  3
        Верхная граница для *_consecutive_ignores.
    laplace_k: float, default  1.0
    laplace_m: float, default 2.0
        Константы для сграживания Laplace-отношений.

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> df = pd.DataFrame({
    ...     "days_of_life": [1, 5, 10],
    ...     "promo_event_count":   [0, 5, 20],
    ...     "promo_visited_count": [0, 3, 15],
    ...     "promo_ignored_count": [0, 2, 5],
    ...     "promo_visit_ratio":   np.nan,
    ...     "promo_ignore_ratio":  np.nan,
    ...     "promo_consecutive_ignores": [0, 2, 4],
    ... })
    >>> pp = EventCountersPostproc(event_name="promo")
    >>> result_df = pp.transform(df)
    >>> print(result_df.round(2).to_string(index=False))
     days_of_life  promo_event_count  promo_visited_count  promo_ignored_count  promo_visit_ratio  promo_ignore_ratio  promo_consecutive_ignores
                1               0.00                 0.00                 0.00               0.33                0.33                          0
                5               0.69                 0.47                 0.34               0.50                0.38                          2
               10               1.10                 0.92                 0.41               0.70                0.26                          3
    """

    event_name: str
    days_col: str = "days_of_life"
    clip_consec: int = 3
    laplace_k: float = 1.0
    laplace_m: float = 2.0

    def fit(self, X, y=None):
        return self

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        p = f"{self.event_name}_"
        days = np.clip(df[self.days_col], 1, None)

        df = df.assign(**{
            f"{p}event_count":   np.log1p(df[f"{p}event_count"]   / days),
            f"{p}visited_count": np.log1p(df[f"{p}visited_count"] / days),
            f"{p}ignored_count": np.log1p(df[f"{p}ignored_count"] / days),

            f"{p}visit_ratio":
                (df[f"{p}visited_count"] + self.laplace_k) /
                (df[f"{p}event_count"]   + self.laplace_k + self.laplace_m),

            f"{p}ignore_ratio":
                (df[f"{p}ignored_count"] + self.laplace_k) /
                (df[f"{p}event_count"]   + self.laplace_k + self.laplace_m),

            f"{p}consecutive_ignores":
                df[f"{p}consecutive_ignores"].clip(upper=self.clip_consec),
        })
        return df



import pandas as pd
from ts_features_sculptor import IntervalEventsMerge

df_main = pd.DataFrame({
    "time": [
        "2024-12-22 01:01:01",
        "2024-12-26 01:01:01",
        "2025-01-01 01:01:01",
    ],
    "value": [6.0, 5.1, 5.5]
})
df_main["time"] = pd.to_datetime(df_main["time"])
df_main.sort_values(by="time", inplace=True)

df_events = pd.DataFrame([])

transformer = IntervalEventsMerge(
    time_col="time",
    events_df=df_events,
    start_col="start",
    end_col="end",
    events_cols=["event_param"],
    fillna=-1.,
    event_name="event"
)
df_result = transformer.transform(df_main)

print("События:")
print(df_events.to_string(index=False))
print("\nРезультат:")
print(df_result.to_string(index=False))


"""Результат работы

События:
Empty DataFrame
Columns: []
Index: []

Результат:
               time  value  event_flag  event_param
2024-12-22 01:01:01    6.0           0         -1.0
2024-12-26 01:01:01    5.1           0         -1.0
2025-01-01 01:01:01    5.5           0         -1.0
"""


import numpy as np
import pandas as pd
from dataclasses import dataclass, field
from sklearn.base import BaseEstimator, TransformerMixin
from .time_validator import TimeValidator


def _td_to_days(delta):
    return delta / np.timedelta64(1, "s") / 86_400.0


@dataclass
class EventDaysFeatures(BaseEstimator, TransformerMixin, TimeValidator):
    """
    Трансформер для вычисления временных характеристик относительно
    интервальных событий.

    Добавляет следующие инженерные признаки:
    - days_to_next_{event_name} дней до начала следующего события
      или fillna если следующее событие отсуствует;
    - days_since_last_{event_name}_end дней от окончания событие
      или fillna;
    - {event_name}_elapsed_days дней от начала события во время
      события или fillna.

    Parameters
    ----------
    time_col: str, default = "time"
        Название колонки с временной меткой.
    events_df: pd.DataFrame,
               default = field(default_factory=pd.DataFrame)
        DataFrame с интервальными событиями.
    start_col: str, default = "start"
        Название колонки с временем начала события в events_df.
    end_col: str, default = "end"
        Название колонки с временем окончания события в events_df.
    event_name: str, default = "event"
        Название события для формирования названий выходных признаков.
    fillna: float, default  = np.nan
        Значения для заполнения выходных характеристик вне зоны их
        действия.

    Examples
    --------
    >>> import pandas as pd
    >>> X = pd.DataFrame({"time": pd.to_datetime([
    ...     "2025-01-01 12:00",   # before promo 1
    ...     "2025-01-02 06:00",   # promo 1 start
    ...     "2025-01-03 12:00",   # promo 1
    ...     "2025-01-04 18:00",   # promo 1 end
    ...     "2025-01-05 00:00",   # after promo 1 before promo 2
    ...     "2025-01-06 12:00",   # after promo 1 before promo 2
    ...     "2025-01-08 06:00",   # promo 2
    ...     "2025-01-10 18:00",   # after promo 2
    ...     "2025-01-15 18:00",   # after promo 2
    ... ])})
    >>> promos = pd.DataFrame({
    ...     "start": pd.to_datetime([
    ...         "2025-01-02 00:00", "2025-01-07 06:00"]),
    ...     "end": pd.to_datetime([
    ...         "2025-01-04 18:00", "2025-01-09 18:00"]),
    ... })
    >>> transformer = EventDaysFeatures(
    ...     time_col="time", events_df=promos, event_name="promo")
    >>> result_df = transformer.transform(X)
    >>> print(result_df.to_string(index=False))
                   time  days_since_last_promo_end  days_to_next_promo  promo_elapsed_days
    2025-01-01 12:00:00                        NaN                0.50                 NaN
    2025-01-02 06:00:00                        NaN                 NaN                0.25
    2025-01-03 12:00:00                        NaN                 NaN                1.50
    2025-01-04 18:00:00                        NaN                 NaN                2.75
    2025-01-05 00:00:00                       0.25                2.25                 NaN
    2025-01-06 12:00:00                       1.75                0.75                 NaN
    2025-01-08 06:00:00                        NaN                 NaN                1.00
    2025-01-10 18:00:00                       1.00                 NaN                 NaN
    2025-01-15 18:00:00                       6.00                 NaN                 NaN
    """

    time_col: str = "time"
    events_df: pd.DataFrame = field(default_factory=pd.DataFrame)
    start_col: str = "start"
    end_col: str = "end"
    event_name: str = "event"
    fillna: float = np.nan

    def fit(self, X: pd.DataFrame, y=None):
        return self

    def _check_events_df(self):
        if self.events_df.empty:
            return
        req = {self.start_col, self.end_col}
        miss = req - set(self.events_df.columns)
        if miss:
            raise ValueError(
                f"EventDaysFeatures: Отсуствуют колонки {miss}")
        bad = (
            self.events_df[self.events_df[self.end_col] <
                           self.events_df[self.start_col]]
        )
        if not bad.empty:
            raise ValueError(
                "EventDaysFeatures: "
                "Найдены события с end < start:\n" + bad.to_string())

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        self._validate_time_column(X)
        self._check_events_df()

        X_ = X.copy()

        _out_cols = [
            f"days_since_last_{self.event_name}_end",
            f"days_to_next_{self.event_name}",
            f"{self.event_name}_elapsed_days",
        ]
        for c in _out_cols:
            X_[c] = self.fillna

        if self.events_df.empty:
            return X_.reset_index(drop=True)

        intervals = self.events_df.copy()
        intervals = intervals.sort_values(
            self.start_col, kind="mergesort"
        ).reset_index(drop=True)
        starts = intervals[self.start_col].values.astype("datetime64[ns]")
        ends = intervals[self.end_col].values.astype("datetime64[ns]")

        ends_sorted = np.sort(ends)

        t = X_[self.time_col].values.astype("datetime64[ns]")
        n = len(t)

        prev_start_idx = np.searchsorted(starts, t, side="right") - 1
        inside = (
            (prev_start_idx >= 0) &
            (t <= ends[np.clip(prev_start_idx, 0, len(ends)-1)])
        )
        outside = ~inside

        if inside.any():
            idx_inside = prev_start_idx[inside]
            elapsed = _td_to_days(t[inside] - starts[idx_inside])
            X_.loc[inside, f"{self.event_name}_elapsed_days"] = elapsed

        next_start_idx = np.searchsorted(starts, t, side="right")
        mask_next = outside & (next_start_idx < len(starts))
        if mask_next.any():
            delta_next = (
                _td_to_days(starts[next_start_idx[mask_next]] -
                t[mask_next])
            )
            X_.loc[mask_next, f"days_to_next_{self.event_name}"] = delta_next

        prev_end_idx = np.searchsorted(ends_sorted, t, side="left") - 1
        mask_prev = outside & (prev_end_idx >= 0)
        if mask_prev.any():
            delta_prev = (
                _td_to_days(t[mask_prev] -
                ends_sorted[prev_end_idx[mask_prev]])
            )
            X_.loc[mask_prev, f"days_since_last_{self.event_name}_end"] = \
                delta_prev

        return X_.reset_index(drop=True)



import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer

from ts_features_sculptor import (

    ToDateTime,
    SortByTime,
    RowLag,
    Tte,
    IntervalEventsMerge,
)

df = pd.DataFrame({
    'time': pd.to_datetime([
        '2025-01-01', '2025-01-05', '2025-01-10',
        '2025-01-21', '2025-01-22', '2025-01-23',
        '2025-02-10', '2025-02-15', '2025-02-25'
    ])
})

promo_df = pd.DataFrame({
    'start': pd.to_datetime(['2025-01-21']),
    'end': pd.to_datetime(['2025-01-23']),
    'intensity': [0.99]
})

n_after = 2
n_before = 3
event_name = 'promo'

pipeline = Pipeline([
    ('to_datetime', ToDateTime(time_col="time")),
    ('sort_by_time', SortByTime(time_col="time")),
    ('tte_builder', Tte(
        time_col="time",
        tte_col='tte'
    )),
    ('promo_merged', IntervalEventsMerge(
        time_col='time',
        events_df=promo_df,
        start_col='start',
        end_col='end',
        event_name='promo',
        events_cols=['intensity'],
        fillna=np.nan
    )),
    # 2) заготовка pre-флага (лид на 1 строку вверх)
    ('pre_lag', RowLag(
        time_col='time',
        feature_col='promo_flag',
        lags=[-n_before],
        fillna=0
    )),
    ('post_lag', RowLag(
        time_col='time',
        feature_col='promo_flag',
        lags=[n_after],
        fillna=0
    )),
    ('flanks', FunctionTransformer(lambda df: (
        df
        .assign(
            **{
                f'{event_name}_pre{n_before}_flag':
                    df[f'{event_name}_flag_rowlag_-{n_before}r']
                    .where(df[f"{event_name}_flag"] == 0, 0)
                    .astype(int),
                f'{event_name}_post{n_after}_flag':
                    df[f'{event_name}_flag_rowlag_{n_after}r']
                    .where(df[f"{event_name}_flag"] == 0, 0)
                    .astype(int),
            }
        )
        .drop(
            columns=[
                f'{event_name}_flag_rowlag_-{n_before}r',
                f'{event_name}_flag_rowlag_{n_after}r'
            ], errors='ignore'
        )
    ))),
])

result_df = pipeline.fit_transform(df)

print(result_df.to_string())




import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline


from ts_features_sculptor import (
    ToDateTime,
    SortByTime,
    Tte,
    IntervalEventsMerge
)

ts_df = pd.DataFrame({
    'time': pd.to_datetime([
        '2024-12-27', '2024-12-28', '2024-12-29',
        '2025-01-01', '2025-01-02', '2025-01-03',
        '2025-01-14', '2025-05-03'
    ])
})

#
long_holiday_df = pd.DataFrame(
{
    'start': pd.to_datetime(['2025-01-01', '2025-05-01']),
    'end': pd.to_datetime(['2025-01-09', '2025-05-12']),
    'category': [1, 2]
})

pipeline = Pipeline([
    ('to_datetime', ToDateTime(time_col="time")),
    ('sort_by_time', SortByTime(time_col="time")),
    ('tte_builder', Tte(
        time_col="time",
        tte_col='tte'
    )),
    ('long_holiday_merged', IntervalEventsMerge(
        time_col='time',
        events_df=long_holiday_df,
        start_col='start',
        end_col='end',
        events_cols=['category'],
        fillna=np.nan
    )),
])

result_df = pipeline.fit_transform(ts_df)

print(result_df.to_string())


import numpy as np
import pandas as pd
from dataclasses import dataclass, field
from sklearn.base import BaseEstimator, TransformerMixin
from .time_validator import TimeValidator


@dataclass
class EventCounters(BaseEstimator, TransformerMixin, TimeValidator):
    """
    Трансформер для вычисления различных счетчиков, связанных с
    интервальными событиями.

    Parameters
    ----------
    time_col: str, default = "time"
         Название колонки с временной меткой.
    events_df: pd.DataFrame, default = pd.DataFrame([])
        DataFrame с интервальными событиями.
    start_col: str, default = "start"
        Название колонки с временем начала события в events_df.
    end_col: str, default = "end"
        Название колонки с временем окончания события в events_df.
    event_name: str, default = "event"
        Название события.
    fillna: float, default = np.nan

    Output features
    ---------------
    - {event_name}_event_count: накопленное число событий
    - {event_name}_visited_count: число посещенных событий
    - {event_name}_ignored_count: число проигнорированных событий
    - {event_name}_visit_ratio: visited_count / event_count
    - {event_name}_ignore_ratio: ignored_count / event_count
    - {event_name}_last_event_visited: флаг посещения последнего события
    - {event_name}_consecutive_ignores: счетчик для последовательно
      проигнорированных событий

    Example
    -------
    >>> import pandas as pd, numpy as np
    >>> df = pd.DataFrame({
    ...     'time': pd.to_datetime(['2025-01-01', '2025-01-10'])})
    >>> events = pd.DataFrame({
    ...     'start': pd.to_datetime(['2025-01-03', '2025-01-08']),
    ...     'end': pd.to_datetime(['2025-01-05', '2025-01-12'])
    ... })
    >>> transformer = EventCounters(
    ...     time_col='time', events_df=events,
    ...     start_col='start', end_col='end',
    ...     event_name='ny', fillna=0.0
    ... )
    >>> result_df = transformer.transform(df)
    >>> print(result_df.to_string(index=False))
          time  ny_event_count  ny_visited_count  ny_ignored_count  ny_visit_ratio  ny_ignore_ratio  ny_last_event_visited  ny_consecutive_ignores
    2025-01-01               0                 0                 0             0.0              0.0                    0.0                       0
    2025-01-10               1                 0                 1             0.0              1.0                    0.0                       1
    """

    time_col:   str = "time"
    events_df:  pd.DataFrame = field(default_factory=pd.DataFrame)
    start_col:  str = "start"
    end_col:    str = "end"
    event_name: str = "event"
    fillna:     float = np.nan

    starts_: np.ndarray = field(init=False, repr=False)
    ends_:   np.ndarray = field(init=False, repr=False)

    _OUT_COLS = [
        "event_count",
        "visited_count",
        "ignored_count",
        "visit_ratio",
        "ignore_ratio",
        "last_event_visited",
        "consecutive_ignores",
    ]

    def _add_empty_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        for col in self._OUT_COLS:
            df[f"{self.event_name}_{col}"] = self.fillna
        return df

    def __post_init__(self):
        if not self.events_df.empty:
            missing = (
                {self.start_col, self.end_col} - set(self.events_df.columns)
            )
            if missing:
                raise ValueError(f"Missing columns in events_df: {missing}")

            ev = self.events_df.sort_values(by=self.start_col) \
                .reset_index(drop=True)
            self.starts_ = ev[self.start_col].to_numpy()
            self.ends_ = ev[self.end_col].to_numpy()
        else:
            self.starts_ = np.array([], dtype="datetime64[ns]")
            self.ends_ = np.array([], dtype="datetime64[ns]")

    def fit(self, X, y=None):
        return self

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self._validate_time_column(df)

        out = df.copy().reset_index(drop=True)
        times = out[self.time_col].to_numpy()
        n = len(out)

        if self.events_df.empty:
            return self._add_empty_columns(out)

        first_time = times.min()
        valid_mask = self.ends_ >= first_time

        if not valid_mask.any():
            return self._add_empty_columns(out)

        starts = self.starts_[valid_mask]
        ends   = self.ends_[valid_mask]

        visited_flags = np.array(
            [((times >= s) & (times <= e)).any()
                 for s, e in zip(starts, ends)],
            dtype=bool,
        )
        cumulative_visited = np.cumsum(visited_flags)

        event_count = np.zeros(n, dtype=int)
        visited_count = np.zeros(n, dtype=int)
        ignored_count = np.zeros(n, dtype=int)
        visit_ratio = np.full(n, self.fillna, dtype=float)
        ignore_ratio = np.full(n, self.fillna, dtype=float)
        last_visited = np.full(n, self.fillna, dtype=float)
        consec_ignores = np.zeros(n, dtype=int)

        run_ignored = 0

        for i in range(n):
            idx = np.searchsorted(ends, times[i], side="right")

            if idx == 0:
                run_ignored = 0
                consec_ignores[i] = 0
                continue

            event_count[i] = idx
            vc = int(cumulative_visited[idx - 1])
            ic = idx - vc

            visited_count[i] = vc
            ignored_count[i] = ic
            visit_ratio[i] = vc / idx
            ignore_ratio[i] = ic / idx
            last_visited[i] = float(visited_flags[idx - 1])

            if visited_flags[idx-1]:
                run_ignored = 0
            else:
                run_ignored += 1
            consec_ignores[i] = run_ignored

        out[f"{self.event_name}_event_count"] = event_count
        out[f"{self.event_name}_visited_count"] = visited_count
        out[f"{self.event_name}_ignored_count"] = ignored_count
        out[f"{self.event_name}_visit_ratio"] = visit_ratio
        out[f"{self.event_name}_ignore_ratio"] = ignore_ratio
        out[f"{self.event_name}_last_event_visited"] = last_visited
        out[f"{self.event_name}_consecutive_ignores"] = consec_ignores

        return out



import numpy as np
import pandas as pd
from dataclasses import dataclass, field
from sklearn.base import BaseEstimator, TransformerMixin
from .time_validator import TimeValidator

@dataclass
class TteEventEffect(BaseEstimator, TransformerMixin, TimeValidator):
    """
    Трансформер вычисляет скользящее среднее tte внутри и вне событий.

    По умолчанию игнорирует последнюю строку в каждом событии,
    так как последнее tte указывает на строку вне события.

    Parameters
    ----------
    time_col: str, default = "time" TODO наследник TimeValidator
        Название колонки базового ряда с временными метками
    tte_col: str, default  = "tte"
        Название колонки с tte (time to event)
    events_df: pd.DataFrame , default = field(default_factory=pd.DataFrame)
        DataFrame с интервальными событиями
    start_col: str, default = "start"
        Название колонки с временем начала события в events_df.
    end_col: str, default = "end"
        Название колонки с временем окончания события в events_df.
    event_name: str, default = "event"
        Название события для формирования имен выходных признаков.
    shift: int, default = 1
        Предотвращаем утечку данных.
    drop_last_in_event: bool, default = True
        Удалять из расчета последнюю строку внутри события.
    fillna: float, default = np.nan
        Заполнитель средних при отсуствии события.
    fallback_uplift: float, default = np.nan
        Заполнитель uplift-признака

    Output features
    ---------------
    {event_name}_inside_tte_avg
    {event_name}_outside_tte_avg
    {event_name}_tte_uplift

    Example
    -------
    >>> import pandas as pd, numpy as np
    >>> df = pd.DataFrame({
    ...     'time': pd.to_datetime([
    ...         '2025-01-01','2025-01-05','2025-01-10']),
    ...     'tte': [4, 5, 6]
    ... })
    >>> promo = pd.DataFrame({
    ...     'start': pd.to_datetime(['2025-01-03']),
    ...     'end': pd.to_datetime(['2025-01-08'])
    ... })
    >>> transformer = TteEventEffect(
    ...      events_df=promo, event_name='promo', shift=1)
    >>> result_df = transformer.transform(df)
    >>> print(result_df.to_string(index=False))
    """

    time_col: str = "time"
    tte_col: str = "tte"
    events_df: pd.DataFrame = field(default_factory=pd.DataFrame)
    start_col: str = "start"
    end_col: str = "end"
    event_name: str = "event"
    shift: int = 1
    drop_last_in_event: bool = True
    fillna: float = np.nan
    fallback_uplift: float = np.nan

    starts_: np.ndarray = field(init=False, repr=False)
    ends_: np.ndarray = field(init=False, repr=False)

    def __post_init__(self):
        if not self.events_df.empty:
            missing = (
                {self.start_col, self.end_col} - set(self.events_df.columns))
            if missing:
                raise ValueError(
                    f"TteEventEffectEstimator: отсуствуют колонки {missing}")
            invalid = (
                self.events_df[self.events_df[self.end_col] <
                self.events_df[self.start_col]]
            )
            if not invalid.empty:
                raise ValueError(
                    f"TteEventEffect: "
                    f"имеются события с end < start:\n{invalid}")
            ev = self.events_df.sort_values(
                by=self.start_col
            ).reset_index(drop=True)
            self.starts_ = ev[self.start_col].to_numpy()
            self.ends_ = ev[self.end_col].to_numpy()
        else:
            self.starts_ = np.array([], dtype="datetime64[ns]")
            self.ends_ = np.array([], dtype="datetime64[ns]")

    def fit(self, X, y=None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        self._validate_time_column(X)
        if self.tte_col not in X.columns:
            raise ValueError(
                f"TteEventEffectEstimator: "
                f"DataFrame не содержит tte_col '{self.tte_col}'")
        X_ = X.copy().reset_index(drop=True)
        n = len(X_)

        # нет событий - просто fillna
        if self.events_df.empty:
            X_[f"{self.event_name}_inside_tte_avg"]  = self.fillna
            X_[f"{self.event_name}_outside_tte_avg"] = self.fillna
            X_[f"{self.event_name}_tte_uplift"]      = self.fillna
            return X_

        # для устранения утечек данных выполняем сдвиг
        tte_orig = X_[self.tte_col].to_numpy(dtype=float)
        tte_shift = np.roll(tte_orig, self.shift)
        if self.shift > 0:
            tte_shift[: self.shift] = np.nan

        times = X_[self.time_col]
        times_np = times.to_numpy()
        interval_end_orig = times + pd.to_timedelta(tte_orig, unit="D")
        interval_end_np = interval_end_orig.to_numpy()

        started = np.searchsorted(self.starts_, times_np, side="right")
        idx = started - 1
        idx[idx < 0] = -1

        ends_per_row = np.full(n, np.datetime64("NaT"), dtype="datetime64[ns]")
        mask_event = idx >= 0
        ends_per_row[mask_event] = self.ends_[idx[mask_event]]

        in_event = mask_event & (times_np <= ends_per_row)
        if self.drop_last_in_event:
            inside_orig = in_event & (interval_end_np <= ends_per_row)
        else:
            inside_orig = in_event

        # compute valid shifted values
        valid = ~np.isnan(tte_shift)

        # shifted classification masks
        inside_mask = np.zeros(n, dtype=bool)
        if self.shift < n:
            inside_mask[self.shift:] = (
                inside_orig[: n - self.shift] & valid[self.shift:]
            )
        outside_mask = valid & ~inside_mask

        inside_sum = np.cumsum(np.where(inside_mask, tte_shift, 0.0))
        inside_cnt = np.cumsum(np.where(inside_mask, 1, 0))
        outside_sum = np.cumsum(np.where(outside_mask, tte_shift, 0.0))
        outside_cnt = np.cumsum(np.where(outside_mask, 1, 0))

        inside_avg = np.divide(
            inside_sum, inside_cnt, where=inside_cnt > 0)
        outside_avg = np.divide(
            outside_sum, outside_cnt, where=outside_cnt > 0)
        inside_avg[inside_cnt == 0] = self.fillna
        outside_avg[outside_cnt == 0] = self.fillna

        uplift = np.divide(
            inside_avg, outside_avg, where=(outside_avg != 0))
        no_data = (inside_cnt == 0) | (outside_cnt == 0) | (outside_avg == 0)
        uplift[no_data] = self.fallback_uplift

        X_[f"{self.event_name}_inside_tte_avg"]  = inside_avg
        X_[f"{self.event_name}_outside_tte_avg"] = outside_avg
        X_[f"{self.event_name}_tte_uplift"]      = uplift

        return X_



import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline


from ts_features_sculptor import (
    ToDateTime,
    SortByTime,
    Tte,
    IntervalEventsMerge,
    TteEventEffect,
    EventCounters,
)

ts_df = pd.DataFrame({
    'time': pd.to_datetime([
        '2024-12-21', '2024-12-23', '2024-12-24',
        '2024-12-26', '2024-12-28', '2024-12-29',
        '2025-01-11', '2025-01-12', '2025-01-14',
    ])
})

ny_df = pd.DataFrame({
    'start': pd.to_datetime(['2025-01-01']),
    'end': pd.to_datetime(['2025-01-10']),
})

event_name = 'promo'

pipeline = Pipeline([
    ('to_datetime', ToDateTime(time_col="time")),
    ('sort_by_time', SortByTime(time_col="time")),
    ('tte_builder', Tte(
        time_col="time",
        tte_col='tte'
    )),
    ('promo_merged', IntervalEventsMerge(
        time_col='time',
        events_df=ny_df,
        start_col='start',
        end_col='end',
        event_name='ny',
        events_cols=[],
        fillna=np.nan
    )),
    ('tte_effect', TteEventEffect(
       events_df=ny_df,
       event_name='ny',
       fallback_uplift=np.nan,
       shift=1
    )),
    ('ny_ignore', EventCounters(
        time_col='time',
        events_df=ny_df,
        start_col='start',
        end_col='end',
        event_name='ny',
        fillna=0
    )),
])

result_df = pipeline.fit_transform(ts_df)

print(result_df.to_string())


import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from dataclasses import dataclass, field
from .time_validator import TimeValidator


@dataclass
class EventDrivenTSCompressor(BaseEstimator, TransformerMixin, TimeValidator):
    """
    Сжимает временной ряд в соответствии с интервальным событийным
    рядом.

    Результирующий временной ряд содержит:
    - признаки последней доступной записи базового временного ряда
      перед началом события;
    - значения из столбцов события (заданные параметром events_cols);
    - количество точек основного ряда, повавшие в интервал [start, end]
      (`<event_name>_count`);
    - лаг (в днях) от последней записи в базовом временном ряде до
      начала события (`<event_name>_lag_days`). То есть, кратчайшее
      расстояние от ближайшей записи базового временного ряда до
      старта события, причем запись в базовом ряде должна быть раньше,
      чем старт события.

    Parameters
    ----------
        time_col, str, default='time':
            Название столбца с метками времени в `X`.
        events_df, pd.DataFrame:
            DataFrame с событиями.
        start_col, str:
            Название столбца начала события в `events_df`.
        end_col, str:
            Название столбца окончания события в `events_df`.
        events_cols, list:
            Список столбцов из `events_df`, которые должны быть
            добавлены в результат.
        events_name, str:
            Базовое имя события для формирования имен выходных колонок
            (`<events_name>_count`, `<events_name>_lag_days`).
        history_end (str or pd.Timestamp, optional):
            Конец доступной истории данных.

    Example
    -------
    >>> X = pd.DataFrame({
    ...     'time': pd.to_datetime([
    ...         '2020-01-01 10:00',
    ...         '2020-01-03 11:00',
    ...         '2020-01-04 12:00',
    ...         '2020-01-06 13:00'
    ...     ]),
    ...     'value': [100, 120, 130, 150]
    ... })
    >>> events_df = pd.DataFrame({
    ...     'start': pd.to_datetime([
    ...         '2020-01-02 16:00', '2020-01-07 19:00']),
    ...     'end': pd.to_datetime([
    ...         '2020-01-05 16:00', '2020-01-08 19:00']),
    ...     'event_type': ['promo1', 'promo2']
    ... })
    >>> transformer = EventDrivenTSCompressor(
    ...     events_df=events_df,
    ...     start_col='start',
    ...     end_col='end',
    ...     time_col='time',
    ...     events_cols=['start', 'event_type'],
    ...     events_name='promo')
    >>> result_df = transformer.transform(X)
    >>> print(result_df.to_string(index=False))
                   time  value               start event_type  promo_count  promo_lag_days
    2020-01-01 10:00:00    100 2020-01-02 16:00:00     promo1          2.0            1.25
    2020-01-06 13:00:00    150 2020-01-07 19:00:00     promo2          NaN            1.25
    """

    events_df: pd.DataFrame
    start_col: str
    end_col: str
    # time_col: str = 'time'
    events_cols: list[str] | None = None
    events_name: str = 'event'
    history_end: pd.Timestamp | None = None

    def fit(self, X, y=None):
        return self

    def transform(self, X_):
        self._validate_time_column(X_)

        X_ = X_.copy()

        events = self.events_df.copy()

        events[self.start_col] = pd.to_datetime(events[self.start_col])
        events[self.end_col] = pd.to_datetime(events[self.end_col])

        times = X_[self.time_col].to_numpy()
        starts = events[self.start_col].to_numpy()
        ends = events[self.end_col].to_numpy()

        idx_last_before = np.searchsorted(times, starts, side='right') - 1

        X_features = X_.iloc[idx_last_before].reset_index(drop=True)

        # Считаем количество записей в интервале [start, end]
        # для каждого события
        idx_start_incl = np.searchsorted(times, starts, side='left')
        idx_end_excl = np.searchsorted(times, ends, side='right')
        counts = idx_end_excl - idx_start_incl
        counts = counts.astype('float64')

        # Определяем границу истории (последнюю доступную дату)
        if self.history_end is not None:
            history_end = pd.to_datetime(self.history_end)
        else:
            history_end = times[-1] if len(times) > 0 else None

        for i in range(len(events)):
            if ends[i] > history_end:
                counts[i] = np.nan

        # дней от последней строки до события

        ## событие принято
        last_times = times[idx_last_before]
        lag_days = (
            (starts - last_times) / np.timedelta64(1,'D')
        )

        lag_days[idx_last_before < 0] = np.nan

        # врезка параметров события
        if self.events_cols is None:
            events_cols_to_use = list(events.columns)
        else:
            events_cols_to_use = [col for col in self.events_cols if
                                  col in events.columns]
        events_subset = events[events_cols_to_use].reset_index(drop=True)

        new_features = pd.DataFrame({
            f"{self.events_name}_count": counts,
            f"{self.events_name}_lag_days": lag_days
        })

        # признаки из последней строки базового ряда,
        # выбранные поля события и новые признаки события
        result_df = pd.concat(
            [X_features, events_subset, new_features],
            axis=1)

        return result_df



import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer

from ts_features_sculptor import (
    ToDateTime,
    SortByTime,
    Tte,
    IntervalEventsMerge,
    TteEventEffect,
)


df = pd.DataFrame({
    'time': pd.to_datetime([
        '2024-12-25', '2025-01-01', '2025-01-05', '2025-01-10',
        '2025-01-21', '2025-01-22', '2025-01-23',
        '2025-02-10', '2025-02-15', '2025-02-25',
        '2025-03-01', '2025-03-03'
    ])
})

promo_df = pd.DataFrame({
    'start': pd.to_datetime(['2025-01-02', '2025-01-21', '2025-02-28']),
    'end': pd.to_datetime(['2025-01-04', '2025-01-23', '2025-03-05']),
    'intensity': [0.1, 0.99, 0.9]
})

event_name = 'promo'

pipeline = Pipeline([
    ('to_datetime', ToDateTime(time_col="time")),
    ('sort_by_time', SortByTime(time_col="time")),
    ('tte_builder', Tte(
        time_col="time",
        tte_col='tte'
    )),
    ('promo_merged', IntervalEventsMerge(
        time_col='time',
        events_df=promo_df,
        start_col='start',
        end_col='end',
        event_name='promo',
        events_cols=['intensity'],
        fillna=np.nan
    )),
    ('promo_effect', TteEventEffect(
        events_df=promo_df,
        event_name="promo",
        # fallback_uplift=999
    )),
])

result_df = pipeline.fit_transform(df)

print(result_df.to_string())

